"""
Beacon to announce via Bonjour (zeroconf)
"""
import atexit
import logging
import select
import time

import salt.utils.beacons
import salt.utils.stringutils

try:
    import pybonjour

    HAS_PYBONJOUR = True
except ImportError:
    HAS_PYBONJOUR = False

log = logging.getLogger(__name__)

__virtualname__ = "bonjour_announce"

LAST_GRAINS = {}
SD_REF = None


def __virtual__():
    if HAS_PYBONJOUR:
        return __virtualname__
    err_msg = "pybonjour library is missing."
    log.error("Unable to load %s beacon: %s", __virtualname__, err_msg)
    return False, err_msg


def _close_sd_ref():
    """
    Close the SD_REF object if it isn't NULL
    For use with atexit.register
    """
    global SD_REF
    if SD_REF:
        SD_REF.close()
        SD_REF = None


def _register_callback(
    sdRef, flags, errorCode, name, regtype, domain
):  # pylint: disable=unused-argument
    if errorCode != pybonjour.kDNSServiceErr_NoError:
        log.error("Bonjour registration failed with error code %s", errorCode)


def validate(config):
    """
    Validate the beacon configuration
    """
    _config = salt.utils.beacons.list_to_dict(config)

    if not isinstance(config, list):
        return False, "Configuration for bonjour_announce beacon must be a list."

    elif not all(x in _config for x in ("servicetype", "port", "txt")):
        return (
            False,
            "Configuration for bonjour_announce beacon must contain servicetype, port and txt items.",
        )
    return True, "Valid beacon configuration."


def _enforce_txt_record_maxlen(key, value):
    """
    Enforces the TXT record maximum length of 255 characters.
    TXT record length includes key, value, and '='.

    :param str key: Key of the TXT record
    :param str value: Value of the TXT record

    :rtype: str
    :return: The value of the TXT record. It may be truncated if it exceeds
             the maximum permitted length. In case of truncation, '...' is
             appended to indicate that the entire value is not present.
    """
    # Add 1 for '=' separator between key and value
    if len(key) + len(value) + 1 > 255:
        # 255 - 3 ('...') - 1 ('=') = 251
        return value[: 251 - len(key)] + "..."
    return value


def beacon(config):
    """
    Broadcast values via zeroconf

    If the announced values are static, it is advised to set run_once: True
    (do not poll) on the beacon configuration.

    The following are required configuration settings:

    - ``servicetype`` - The service type to announce
    - ``port`` - The port of the service to announce
    - ``txt`` - The TXT record of the service being announced as a dict. Grains
      can be used to define TXT values using one of following two formats:

      - ``grains.<grain_name>``
      - ``grains.<grain_name>[i]`` where i is an integer representing the
        index of the grain to use. If the grain is not a list, the index is
        ignored.

    The following are optional configuration settings:

    - ``servicename`` - Set the name of the service. Will use the hostname from
      the minion's ``host`` grain if this value is not set.
    - ``reset_on_change`` - If ``True`` and there is a change in TXT records
      detected, it will stop announcing the service and then restart announcing
      the service. This interruption in service announcement may be desirable
      if the client relies on changes in the browse records to update its cache
      of TXT records. Defaults to ``False``.
    - ``reset_wait`` - The number of seconds to wait after announcement stops
      announcing and before it restarts announcing in the case where there is a
      change in TXT records detected and ``reset_on_change`` is ``True``.
      Defaults to ``0``.
    - ``copy_grains`` - If ``True``, Salt will copy the grains passed into the
      beacon when it backs them up to check for changes on the next iteration.
      Normally, instead of copy, it would use straight value assignment. This
      will allow detection of changes to grains where the grains are modified
      in-place instead of completely replaced.  In-place grains changes are not
      currently done in the main Salt code but may be done due to a custom
      plug-in. Defaults to ``False``.

    Example Config

    .. code-block:: yaml

       beacons:
         bonjour_announce:
           - run_once: True
           - servicetype: _demo._tcp
           - port: 1234
           - txt:
               ProdName: grains.productname
               SerialNo: grains.serialnumber
               Comments: 'this is a test'
    """
    ret = []
    changes = {}
    txt = {}

    global LAST_GRAINS
    global SD_REF

    config = salt.utils.beacons.list_to_dict(config)

    if "servicename" in config:
        servicename = config["servicename"]
    else:
        servicename = __grains__["host"]
        # Check for hostname change
        if LAST_GRAINS and LAST_GRAINS["host"] != servicename:
            changes["servicename"] = servicename

    if LAST_GRAINS and config.get("reset_on_change", False):
        # Check for IP address change in the case when we reset on change
        if LAST_GRAINS.get("ipv4", []) != __grains__.get("ipv4", []):
            changes["ipv4"] = __grains__.get("ipv4", [])
        if LAST_GRAINS.get("ipv6", []) != __grains__.get("ipv6", []):
            changes["ipv6"] = __grains__.get("ipv6", [])

    for item in config["txt"]:
        changes_key = "txt." + salt.utils.stringutils.to_unicode(item)
        if config["txt"][item].startswith("grains."):
            grain = config["txt"][item][7:]
            grain_index = None
            square_bracket = grain.find("[")
            if square_bracket != -1 and grain[-1] == "]":
                grain_index = int(grain[square_bracket + 1 : -1])
                grain = grain[:square_bracket]

            grain_value = __grains__.get(grain, "")
            if isinstance(grain_value, list):
                if grain_index is not None:
                    grain_value = grain_value[grain_index]
                else:
                    grain_value = ",".join(grain_value)
            txt[item] = _enforce_txt_record_maxlen(item, grain_value)
            if LAST_GRAINS and (
                LAST_GRAINS.get(grain, "") != __grains__.get(grain, "")
            ):
                changes[changes_key] = txt[item]
        else:
            txt[item] = _enforce_txt_record_maxlen(item, config["txt"][item])

        if not LAST_GRAINS:
            changes[changes_key] = txt[item]

    if changes:
        txt_record = pybonjour.TXTRecord(items=txt)
        if not LAST_GRAINS:
            changes["servicename"] = servicename
            changes["servicetype"] = config["servicetype"]
            changes["port"] = config["port"]
            changes["ipv4"] = __grains__.get("ipv4", [])
            changes["ipv6"] = __grains__.get("ipv6", [])
            SD_REF = pybonjour.DNSServiceRegister(
                name=servicename,
                regtype=config["servicetype"],
                port=config["port"],
                txtRecord=txt_record,
                callBack=_register_callback,
            )
            atexit.register(_close_sd_ref)
            ready = select.select([SD_REF], [], [])
            if SD_REF in ready[0]:
                pybonjour.DNSServiceProcessResult(SD_REF)
        elif config.get("reset_on_change", False) or "servicename" in changes:
            # A change in 'servicename' requires a reset because we can only
            # directly update TXT records
            SD_REF.close()
            SD_REF = None
            reset_wait = config.get("reset_wait", 0)
            if reset_wait > 0:
                time.sleep(reset_wait)
            SD_REF = pybonjour.DNSServiceRegister(
                name=servicename,
                regtype=config["servicetype"],
                port=config["port"],
                txtRecord=txt_record,
                callBack=_register_callback,
            )
            ready = select.select([SD_REF], [], [])
            if SD_REF in ready[0]:
                pybonjour.DNSServiceProcessResult(SD_REF)
        else:
            txt_record_raw = str(txt_record).encode("utf-8")
            pybonjour.DNSServiceUpdateRecord(
                SD_REF, RecordRef=None, flags=0, rdata=txt_record_raw
            )

        ret.append({"tag": "result", "changes": changes})

    if config.get("copy_grains", False):
        LAST_GRAINS = __grains__.copy()
    else:
        LAST_GRAINS = __grains__

    return ret



import logging

import pytest

import salt.modules.tls as tls
from salt.utils.versions import Version
from tests.support.helpers import SKIP_INITIAL_PHOTONOS_FAILURES
from tests.support.mock import MagicMock, mock_open, patch

log = logging.getLogger(__name__)


pytestmark = [
    SKIP_INITIAL_PHOTONOS_FAILURES,
]


@pytest.fixture
def configure_loader_modules():
    return {tls: {}}


@pytest.fixture(scope="module")
def tls_test_data():
    return {
        "create_ca": {
            "bits": 2048,
            "CN": "localhost",
            "C": "US",
            "ST": "Utah",
            "L": "Salt Lake City",
            "O": "SaltStack",
            "OU": "Test Unit",
            "emailAddress": "xyz@pdq.net",
            "digest": "sha256",
            "replace": False,
        },
        "ca_signed_cert": "",
        "ca_signed_key": "",
        "ca_cert": """-----BEGIN CERTIFICATE-----
MIIEejCCA2KgAwIBAgIRANW6IUG5rMez0vPi3cSmS3QwDQYJKoZIhvcNAQELBQAw
eTELMAkGA1UEBhMCVVMxDTALBgNVBAgMBFV0YWgxFzAVBgNVBAcMDlNhbHQgTGFr
ZSBDaXR5MRIwEAYDVQQKDAlTYWx0U3RhY2sxEjAQBgNVBAMMCWxvY2FsaG9zdDEa
MBgGCSqGSIb3DQEJARYLeHl6QHBkcS5uZXQwHhcNMTUwNTA1MTYzOTIxWhcNMTYw
NTA0MTYzOTIxWjB5MQswCQYDVQQGEwJVUzENMAsGA1UECAwEVXRhaDEXMBUGA1UE
BwwOU2FsdCBMYWtlIENpdHkxEjAQBgNVBAoMCVNhbHRTdGFjazESMBAGA1UEAwwJ
bG9jYWxob3N0MRowGAYJKoZIhvcNAQkBFgt4eXpAcGRxLm5ldDCCASIwDQYJKoZI
hvcNAQEBBQADggEPADCCAQoCggEBAMNvHc8LwpI5/NiwRTWYG34WQ5vau8gkj+8p
5KehXDNmDcCY8QW9xNaCxY6Atg2Dwh5vEacubKRcnQL9SFKYHa4ddtnkISzSkdZN
ImY7ZVQteDIVNJmy7DrZ4RvWTr2ezXYLv8oNkqrKhynt5xIBXZWslWUav1pOp8z8
N+LeXaASVyajqB5TiN8HJR/up9MlSfy/zhtm6x6SIUsEZa+zK7m06/Glrr4WZFOV
LbOwxl36JpjywWTNcrXJd052U/377tUATXpepALBUUOIvWeGF7mrSTZkdhqRZRTe
Jr2+48zIuyMeB+JlY4UpR04pQNqstHimkyjxFfN/TKFqlhYqYjkCAwEAAaOB/DCB
+TASBgNVHRMBAf8ECDAGAQH/AgEAMA4GA1UdDwEB/wQEAwIBBjAdBgNVHQ4EFgQU
WBvk3qjnltkxKtEQxqYn5+KwYWkwgbMGA1UdIwSBqzCBqIAUWBvk3qjnltkxKtEQ
xqYn5+KwYWmhfaR7MHkxCzAJBgNVBAYTAlVTMQ0wCwYDVQQIDARVdGFoMRcwFQYD
VQQHDA5TYWx0IExha2UgQ2l0eTESMBAGA1UECgwJU2FsdFN0YWNrMRIwEAYDVQQD
DAlsb2NhbGhvc3QxGjAYBgkqhkiG9w0BCQEWC3h5ekBwZHEubmV0ghEA1bohQbms
x7PS8+LdxKZLdDANBgkqhkiG9w0BAQsFAAOCAQEALe312Oe8e+VjhnItcjQFuwcP
TaLf3+DTWaQLU1C8H78E75WE9UiRiVCyTpOLt/nONFkIKE275nCLPGCXn5JTZYVB
CxGFTRqnQ+8bdhZA6LYQPXieGikjTy+P2oiKOvPnYsATUXLbZ3ee+zEgBFGbbxNX
Argd3Vahg7Onu3ynsJz9a+hmwVqTX70Ykrrm+b/YtwKPfHeXTMxkX23jc4R7D+ED
VvROFJ27hFPLVaJrsq3EHb8ZkQHmRCzK2sMIPyJb2e0BOKDEZEphNxiIjerDnH7n
xDSMW0jK9FEv/W/sSBwoEolh3Q2e0gfd2vo7bEGvNF4eTqFsoAfeiAjWk65q0Q==
-----END CERTIFICATE-----""",
        "ca_cert_key": """-----BEGIN PRIVATE KEY-----
MIIEvQIBADANBgkqhkiG9w0BAQEFAASCBKcwggSjAgEAAoIBAQDDbx3PC8KSOfzY
sEU1mBt+FkOb2rvIJI/vKeSnoVwzZg3AmPEFvcTWgsWOgLYNg8IebxGnLmykXJ0C
/UhSmB2uHXbZ5CEs0pHWTSJmO2VULXgyFTSZsuw62eEb1k69ns12C7/KDZKqyocp
7ecSAV2VrJVlGr9aTqfM/Dfi3l2gElcmo6geU4jfByUf7qfTJUn8v84bZusekiFL
BGWvsyu5tOvxpa6+FmRTlS2zsMZd+iaY8sFkzXK1yXdOdlP9++7VAE16XqQCwVFD
iL1nhhe5q0k2ZHYakWUU3ia9vuPMyLsjHgfiZWOFKUdOKUDarLR4ppMo8RXzf0yh
apYWKmI5AgMBAAECggEBAIRqIRRLr4VL7NkUdZAeg2Imy6Apz9mHjE5LYDWDyui4
WNEJzyRIs7lz2U74PmFhyIC+WIOhnNKwPWHtIrdzgYibRg/T1fZ8pXtBv/DshXdH
Z4zneUA6TnyBa1hlF+y6UBOPWl8YWyuFFZd/LXSxoCrtSDu8p7IUYPUuXt9EMsNk
+rQhBZH7iiOx2M7ckLb/gG4tElqh/QQ0Qhuy1e80oMfO9w1XxIThETx0LxQ7Uad5
vk5BQoEI+n/W1tPhXwQ/hndn3ub3pgfpKXIKi7AdySsdxTda4pJLyEwAm9MHrMNm
XvlBdQ+/vq4NSRk2oVVFzJ1yX7bfmrw+6ZqY36oOrUECgYEA4KvnSgeUyG4ouwA9
zTehvFhEWfFyj/y8Tl2OcaIeCvHahN3n6n++Rx25JYrQDhguMGgMHJ0dcZr0Wvvc
EPOAxjam96p2HJMYdpYmrwpguHDu4Jk+fOqgc8Ms541P3+sMej1UQxRaQ6c/g1EM
Pl0g2GdDN4bw60nbEwOXp0qFI1UCgYEA3q+GJDA+gkXcMqKN06VEdjPvm+BpOygx
VvO0cG6njcmVjJKPePIYbjcBHduQRbCGxLmntLOyVg6Ign1+r/LaG7AADYXMdURD
cn4J9ANKuT1lZcs/JRatA1qigwlk8E7vIqq0E/1EiEbRWVE34lgN3dkUOqAXLBfi
9p2aJMOtC1UCgYAKnDOtDFSbbpBf3HAOu/zYXzbDJKLrZ90gukxa03Qlwiw2sCAe
s++xfhbbTgXrVHsB8Df6NfVJAy9dCJ3o8wb21WfnNFalnNC/8PFcvNm6fCLb2oDX
92Ciduos+UB3a6tILpNHI7Prk/9s3Sv92foOHjpPagEAq5k7+aR00xEcjQKBgEhn
r+kCWsDG8ELygcToPqtkVatMO0sF1Y0dLnVENWyvt9V+LfI4XWMwtUc9Bdry+87p
QrNJnlnG3fH31gJlpy9Leajr8T/L01ZdzuStUVWLtfV0MXLgvZ6SkLakjlJoh+6w
rF63gdoBlL5C3zXURaX1mFM7jG1E0wI22lDL4u8FAoGAaeghxaW4V0keS0KpdPvL
kmEhQ5VDAG92fmzQa0wkBoTbsS4kitHj5QTfNJgdCDJbZ7srkMbQR7oVGeZKIDBk
L7lyhyrtHaR/r/fdUNEJAVsQt3Shaa5c5srLh6vzGcBsV7/nQqrEP+GadvGSbHNT
bymYbi0l2pWqQLA2sPoRHNw=
-----END PRIVATE KEY-----""",
    }


@pytest.mark.skip_on_windows(reason="Skipping on Windows per Shane's suggestion")
def test_create_ca_permissions_on_cert_and_key(tmp_path, tls_test_data):
    ca_name = "test_ca"
    certp = tmp_path / ca_name / "{}_ca_cert.crt".format(ca_name)
    certk = tmp_path / ca_name / "{}_ca_cert.key".format(ca_name)
    mock_opt = MagicMock(return_value=str(tmp_path))
    mock_ret = MagicMock(return_value=0)

    with patch.dict(
        tls.__salt__, {"config.option": mock_opt, "cmd.retcode": mock_ret}
    ), patch.dict(tls.__opts__, {"hash_type": "sha256", "cachedir": str(tmp_path)}):
        tls.create_ca(ca_name, days=365, fixmode=False, **tls_test_data["create_ca"])
        assert certp.stat().st_mode & 0o7777 == 0o644
        assert certk.stat().st_mode & 0o7777 == 0o600


@pytest.mark.skip_on_windows(reason="Skipping on Windows per Shane's suggestion")
def test_create_csr_permissions_on_csr_and_key(tmp_path, tls_test_data):
    ca_name = "test_ca"
    csrp = (
        tmp_path / ca_name / "certs" / "{}.csr".format(tls_test_data["create_ca"]["CN"])
    )
    keyp = (
        tmp_path / ca_name / "certs" / "{}.key".format(tls_test_data["create_ca"]["CN"])
    )

    mock_opt = MagicMock(return_value=str(tmp_path))
    mock_ret = MagicMock(return_value=0)
    mock_pgt = MagicMock(return_value=False)

    with patch.dict(
        tls.__salt__,
        {"config.option": mock_opt, "cmd.retcode": mock_ret, "pillar.get": mock_pgt},
    ), patch.dict(tls.__opts__, {"hash_type": "sha256", "cachedir": str(tmp_path)}):
        ca_ret = tls.create_ca(ca_name, days=365, **tls_test_data["create_ca"])
        assert ca_ret
        csr_ret = tls.create_csr(ca_name, **tls_test_data["create_ca"])
        assert csr_ret

        assert csrp.exists()
        assert keyp.exists()

        assert csrp.stat().st_mode & 0o7777 == 0o644
        assert keyp.stat().st_mode & 0o7777 == 0o600


@pytest.mark.skip_on_windows(reason="Skipping on Windows per Shane's suggestion")
def test_create_self_signed_cert_permissions_on_csr_cert_and_key(
    tmp_path, tls_test_data
):
    ca_name = "test_ca"
    certp = (
        tmp_path / ca_name / "certs" / "{}.crt".format(tls_test_data["create_ca"]["CN"])
    )
    keyp = (
        tmp_path / ca_name / "certs" / "{}.key".format(tls_test_data["create_ca"]["CN"])
    )

    mock_opt = MagicMock(return_value=str(tmp_path))
    mock_ret = MagicMock(return_value=0)
    mock_pgt = MagicMock(return_value=False)

    with patch.dict(
        tls.__salt__,
        {"config.option": mock_opt, "cmd.retcode": mock_ret, "pillar.get": mock_pgt},
    ), patch.dict(tls.__opts__, {"hash_type": "sha256", "cachedir": str(tmp_path)}):
        tls.create_self_signed_cert(ca_name, days=365, **tls_test_data["create_ca"])

        assert certp.stat().st_mode & 0o7777 == 0o644
        assert keyp.stat().st_mode & 0o7777 == 0o600


@pytest.mark.parametrize(
    "expected_path, ca_path, config_option",
    (
        # Nothing passed in, nothing in config
        (None, None, {}),
        # fnord passed in so fnord should come out. Probably? This was based on
        # the existing code when the test was written.
        ("fnord", "fnord", {}),
        # same as above - config should be ignored.
        (
            "fnord",
            "fnord",
            {
                "ca.contextual_cert_base_path": "not fnord",
                "ca.cert_base_path": "also not fnord",
            },
        ),
        # Nothing passed in, look in contextual_cert_base_path first
        (
            "pick me",
            None,
            {
                "ca.contextual_cert_base_path": "pick me",
                "ca.cert_base_path": "also not fnord",
            },
        ),
        # Nothing passed in, or in contextual_cert_base_path first, fallback to cert_base_path
        (
            "now pick me!",
            None,
            {"ca.contextual_cert_base_path": "", "ca.cert_base_path": "now pick me!"},
        ),
        (
            "me too",
            None,
            {"ca.contextual_cert_base_path": None, "ca.cert_base_path": "me too"},
        ),
        ("me three", None, {"ca.cert_base_path": "me three"}),
    ),
)
def test_cert_base_path(expected_path, ca_path, config_option):
    """
    Test for retrieving cert base path
    """

    def path_getter(key):
        return config_option.get(key)

    mock_opt = MagicMock(side_effect=path_getter)
    with patch.dict(tls.__salt__, {"config.option": mock_opt}):
        assert tls.cert_base_path(ca_path) == expected_path


def test_when_ca_path_not_provided_and_no_contextual_cert_base_path_should_be_context_cert_base_path():
    # It's unclear on whether or not this is necessary. salt.modules.tls
    # doesn't have a specific way to set the context value, but it's possible
    # that it's getting set out of band somewhere. This at least ensures that
    # the behavior doesn't unexpectedly change, but it may be useless.
    with patch.dict(
        tls.__salt__, {"config.option": MagicMock(side_effect=(None, "blorp"))}
    ), patch.dict(tls.__context__, {"ca.cert_base_path": "expected path"}):
        actual_path = tls.cert_base_path()
        assert actual_path == "expected path"


@pytest.mark.parametrize(
    "expected_path, new_path, context",
    (
        (None, "", {}),
        (None, None, {}),
        ("fnord", None, {"ca.cert_base_path": "fnord"}),
        (
            "fnord",
            None,
            {"ca.contextual_cert_base_path": "fnord", "ca.cert_base_path": "not fnord"},
        ),
        (
            "fnord",
            "fnord",
            {
                "ca.contextual_cert_base_path": "not fnord",
                "ca.cert_base_path": "not fnord",
            },
        ),
    ),
)
def test_set_ca_cert_path(expected_path, new_path, context):
    """
    Test for setting the cert base path
    """
    with patch.dict(tls.__context__, context), patch.dict(
        tls.__salt__, {"config.option": MagicMock(return_value=None)}
    ):
        tls.set_ca_path(new_path)
        actual_path = tls.cert_base_path()
        assert actual_path == expected_path


@pytest.mark.parametrize("exists", [True, False])
def test_ca_exists(exists):
    """
    Test to see if ca does not exist
    """
    ca_path = "/tmp/test_tls"
    ca_name = "test_ca"
    mock_opt = MagicMock(return_value=ca_path)
    with patch.dict(tls.__salt__, {"config.option": mock_opt}), patch(
        "os.path.exists", MagicMock(return_value=exists)
    ), patch("salt.modules.tls.maybe_fix_ssl_version", MagicMock(return_value=exists)):
        assert tls.ca_exists(ca_name) is exists


def test_get_ca_fail():
    """
    Test get_ca failure
    """
    ca_path = "/tmp/test_tls"
    ca_name = "test_ca"
    mock_opt = MagicMock(return_value=ca_path)
    with patch.dict(tls.__salt__, {"config.option": mock_opt}), patch(
        "os.path.exists", MagicMock(return_value=False)
    ), patch("salt.modules.tls.maybe_fix_ssl_version", MagicMock(return_value=True)):
        with pytest.raises(ValueError):
            tls.get_ca(ca_name)


def test_get_ca_text(tls_test_data):
    """
    Test get_ca text
    """
    ca_path = "/tmp/test_tls"
    ca_name = "test_ca"
    mock_opt = MagicMock(return_value=ca_path)
    with patch(
        "salt.utils.files.fopen", mock_open(read_data=tls_test_data["ca_cert"])
    ), patch.dict(tls.__salt__, {"config.option": mock_opt}), patch(
        "os.path.exists", MagicMock(return_value=True)
    ), patch(
        "salt.modules.tls.maybe_fix_ssl_version", MagicMock(return_value=True)
    ):
        assert tls.get_ca(ca_name, as_text=True) == tls_test_data["ca_cert"]


def test_get_ca():
    """
    Test get_ca
    """
    ca_path = "/tmp/test_tls"
    ca_name = "test_ca"
    certp = "{0}/{1}/{1}_ca_cert.crt".format(ca_path, ca_name)
    mock_opt = MagicMock(return_value=ca_path)
    with patch.dict(tls.__salt__, {"config.option": mock_opt}), patch(
        "os.path.exists", MagicMock(return_value=True)
    ), patch("salt.modules.tls.maybe_fix_ssl_version", MagicMock(return_value=True)):
        assert tls.get_ca(ca_name) == certp


def test_cert_info(tls_test_data):
    """
    Test cert info
    """
    with patch("os.path.exists", MagicMock(return_value=True)), patch(
        "salt.modules.tls.maybe_fix_ssl_version", MagicMock(return_value=True)
    ):
        ca_path = "/tmp/test_tls"
        ca_name = "test_ca"
        certp = "{0}/{1}/{1}_ca_cert.crt".format(ca_path, ca_name)
        ret = {
            "not_after": 1462379961,
            "signature_algorithm": "sha256WithRSAEncryption",
            "extensions": None,
            "fingerprint": (
                "96:72:B3:0A:1D:34:37:05:75:57:44:7E:08:81:A7:09:"
                "0C:E1:8F:5F:4D:0C:49:CE:5B:D2:6B:45:D3:4D:FF:31"
            ),
            "serial_number": 284092004844685647925744086791559203700,
            "subject": {
                "C": "US",
                "CN": "localhost",
                "L": "Salt Lake City",
                "O": "SaltStack",
                "ST": "Utah",
                "emailAddress": "xyz@pdq.net",
            },
            "not_before": 1430843961,
            "issuer": {
                "C": "US",
                "CN": "localhost",
                "L": "Salt Lake City",
                "O": "SaltStack",
                "ST": "Utah",
                "emailAddress": "xyz@pdq.net",
            },
        }

        def ignore_extensions(data):
            """
            Ignore extensions pending a resolution of issue 24338
            """
            if "extensions" in data.keys():
                data["extensions"] = None
            return data

        # older pyopenssl versions don't have extensions or
        # signature_algorithms
        def remove_not_in_result(source, reference):
            if "signature_algorithm" not in reference:
                del source["signature_algorithm"]
            if "extensions" not in reference:
                del source["extensions"]

        with patch(
            "salt.utils.files.fopen", mock_open(read_data=tls_test_data["ca_cert"])
        ):
            try:
                result = ignore_extensions(tls.cert_info(certp))
            except AttributeError as err:
                # PyOpenSSL version 16.0.0 has an upstream bug in it where a call is made
                # in OpenSSL/crypto.py in the get_signature_algorithm function referencing
                # the cert_info attribute, which doesn't exist. This was fixed in subsequent
                # releases of PyOpenSSL with https://github.com/pyca/pyopenssl/pull/476
                if (
                    "'_cffi_backend.CDataGCP' object has no attribute 'cert_info'"
                    == str(err)
                ):
                    log.exception(err)
                    pytest.skip(
                        "Encountered an upstream error with PyOpenSSL: {}".format(err)
                    )
                if "'_cffi_backend.CDataGCP' object has no attribute 'object'" == str(
                    err
                ):
                    log.exception(err)
                    pytest.skip(
                        "Encountered an upstream error with PyOpenSSL: {}".format(err)
                    )
                # python-openssl version 0.14, when installed with the "junos-eznc" pip
                # package, causes an error on this test. Newer versions of PyOpenSSL do not have
                # this issue. If 0.14 is installed and we hit this error, skip the test.
                OpenSSL = pytest.importorskip("OpenSSL")
                if Version(OpenSSL.__version__) == Version("0.14"):
                    log.exception(err)
                    pytest.skip(
                        "Encountered a package conflict. OpenSSL version 0.14"
                        ' cannot be used with the "junos-eznc" pip package on this'
                        " test. Skipping."
                    )
                result = {}

        remove_not_in_result(ret, result)
        assert result == ret


def test_create_ca(tmp_path, tls_test_data):
    """
    Test creating CA cert
    """
    ca_path = tmp_path
    ca_name = "test_ca"
    certp = "{0}/{1}/{1}_ca_cert.crt".format(ca_path, ca_name)
    certk = "{0}/{1}/{1}_ca_cert.key".format(ca_path, ca_name)
    ret = 'Created Private Key: "{}" Created CA "{}": "{}"'.format(
        certk, ca_name, certp
    )
    mock_opt = MagicMock(return_value=ca_path)
    mock_ret = MagicMock(return_value=0)
    with patch.dict(
        tls.__salt__, {"config.option": mock_opt, "cmd.retcode": mock_ret}
    ), patch.dict(tls.__opts__, {"hash_type": "sha256", "cachedir": ca_path}), patch(
        "salt.modules.tls.maybe_fix_ssl_version", MagicMock(return_value=True)
    ):
        assert (
            tls.create_ca(
                ca_name, days=365, fixmode=False, **tls_test_data["create_ca"]
            )
            == ret
        )


def test_recreate_ca(tmp_path, tls_test_data):
    """
    Test creating CA cert when one already exists
    """
    ca_path = tmp_path
    ca_name = "test_ca"
    certp = "{0}/{1}/{1}_ca_cert.crt".format(ca_path, ca_name)
    certk = "{0}/{1}/{1}_ca_cert.key".format(ca_path, ca_name)
    ret = f'Created Private Key: "{certk}" Created CA "{ca_name}": "{certp}"'
    mock_opt = MagicMock(return_value=ca_path)
    mock_ret = MagicMock(return_value=0)
    with patch(
        "salt.modules.tls.maybe_fix_ssl_version", MagicMock(return_value=True)
    ), patch.dict(
        tls.__salt__, {"config.option": mock_opt, "cmd.retcode": mock_ret}
    ), patch.dict(
        tls.__opts__, {"hash_type": "sha256", "cachedir": ca_path}
    ), patch.dict(
        tls_test_data["create_ca"], {"replace": True}
    ):
        tls.create_ca(ca_name)
        assert (
            tls.create_ca(
                ca_name, days=365, fixmode=False, **tls_test_data["create_ca"]
            )
            == ret
        )


def test_create_csr(tmp_path, tls_test_data):
    """
    Test creating certificate signing request
    """
    ca_path = tmp_path
    ca_name = "test_ca"
    certp = "{}/{}/certs/{}.csr".format(
        ca_path, ca_name, tls_test_data["create_ca"]["CN"]
    )
    certk = "{}/{}/certs/{}.key".format(
        ca_path, ca_name, tls_test_data["create_ca"]["CN"]
    )
    ret = 'Created Private Key: "{}" Created CSR for "{}": "{}"'.format(
        certk, tls_test_data["create_ca"]["CN"], certp
    )
    mock_opt = MagicMock(return_value=ca_path)
    mock_ret = MagicMock(return_value=0)
    mock_pgt = MagicMock(return_value=False)
    with patch.dict(
        tls.__salt__,
        {
            "config.option": mock_opt,
            "cmd.retcode": mock_ret,
            "pillar.get": mock_pgt,
        },
    ), patch.dict(tls.__opts__, {"hash_type": "sha256", "cachedir": ca_path}), patch(
        "salt.modules.tls.maybe_fix_ssl_version", MagicMock(return_value=True)
    ):
        tls.create_ca(ca_name)
        assert tls.create_csr(ca_name, **tls_test_data["create_ca"]) == ret


def test_recreate_csr(tmp_path, tls_test_data):
    """
    Test creating certificate signing request when one already exists
    """
    ca_path = tmp_path
    ca_name = "test_ca"
    certp = "{}/{}/certs/{}.csr".format(
        ca_path, ca_name, tls_test_data["create_ca"]["CN"]
    )
    certk = "{}/{}/certs/{}.key".format(
        ca_path, ca_name, tls_test_data["create_ca"]["CN"]
    )
    ret = 'Created Private Key: "{}" Created CSR for "{}": "{}"'.format(
        certk, tls_test_data["create_ca"]["CN"], certp
    )
    mock_opt = MagicMock(return_value=ca_path)
    mock_ret = MagicMock(return_value=0)
    mock_pgt = MagicMock(return_value=False)
    with patch.dict(
        tls.__salt__,
        {
            "config.option": mock_opt,
            "cmd.retcode": mock_ret,
            "pillar.get": mock_pgt,
        },
    ), patch.dict(
        tls.__opts__, {"hash_type": "sha256", "cachedir": ca_path}
    ), patch.dict(
        tls_test_data["create_ca"], {"replace": True}
    ), patch(
        "salt.modules.tls.maybe_fix_ssl_version", MagicMock(return_value=True)
    ):
        tls.create_ca(ca_name)
        tls.create_csr(ca_name)
        assert tls.create_csr(ca_name, **tls_test_data["create_ca"]) == ret


def test_create_self_signed_cert(tmp_path, tls_test_data):
    """
    Test creating self signed certificate
    """
    ca_path = tmp_path
    tls_dir = "test_tls"
    certp = "{}/{}/certs/{}.crt".format(
        ca_path, tls_dir, tls_test_data["create_ca"]["CN"]
    )
    certk = "{}/{}/certs/{}.key".format(
        ca_path, tls_dir, tls_test_data["create_ca"]["CN"]
    )
    ret = 'Created Private Key: "{}" Created Certificate: "{}"'.format(certk, certp)
    mock_opt = MagicMock(return_value=ca_path)
    with patch.dict(tls.__salt__, {"config.option": mock_opt}), patch.dict(
        tls.__opts__, {"hash_type": "sha256", "cachedir": ca_path}
    ), patch("salt.modules.tls.maybe_fix_ssl_version", MagicMock(return_value=True)):
        assert (
            tls.create_self_signed_cert(
                tls_dir=tls_dir, days=365, **tls_test_data["create_ca"]
            )
            == ret
        )


def test_recreate_self_signed_cert(tmp_path, tls_test_data):
    """
    Test creating self signed certificate when one already exists
    """
    ca_path = tmp_path
    tls_dir = "test_tls"
    certp = "{}/{}/certs/{}.crt".format(
        ca_path, tls_dir, tls_test_data["create_ca"]["CN"]
    )
    certk = "{}/{}/certs/{}.key".format(
        ca_path, tls_dir, tls_test_data["create_ca"]["CN"]
    )
    ret = 'Created Private Key: "{}" Created Certificate: "{}"'.format(certk, certp)
    mock_opt = MagicMock(return_value=ca_path)
    with patch.dict(tls.__salt__, {"config.option": mock_opt}), patch.dict(
        tls.__opts__, {"hash_type": "sha256", "cachedir": ca_path}
    ), patch("salt.modules.tls.maybe_fix_ssl_version", MagicMock(return_value=True)):
        assert (
            tls.create_self_signed_cert(
                tls_dir=tls_dir, days=365, **tls_test_data["create_ca"]
            )
            == ret
        )


def test_create_ca_signed_cert(tmp_path, tls_test_data):
    """
    Test signing certificate from request
    """
    ca_path = tmp_path
    ca_name = "test_ca"
    certp = "{}/{}/certs/{}.crt".format(
        ca_path, ca_name, tls_test_data["create_ca"]["CN"]
    )
    ret = 'Created Certificate for "{}": "{}"'.format(
        tls_test_data["create_ca"]["CN"], certp
    )
    mock_opt = MagicMock(return_value=ca_path)
    mock_ret = MagicMock(return_value=0)
    mock_pgt = MagicMock(return_value=False)
    with patch.dict(
        tls.__salt__,
        {
            "config.option": mock_opt,
            "cmd.retcode": mock_ret,
            "pillar.get": mock_pgt,
        },
    ), patch.dict(tls.__opts__, {"hash_type": "sha256", "cachedir": ca_path}), patch(
        "salt.modules.tls.maybe_fix_ssl_version", MagicMock(return_value=True)
    ):
        tls.create_ca(ca_name)
        tls.create_csr(ca_name, **tls_test_data["create_ca"])
        assert (
            tls.create_ca_signed_cert(ca_name, tls_test_data["create_ca"]["CN"]) == ret
        )


def test_recreate_ca_signed_cert(tmp_path, tls_test_data):
    """
    Test signing certificate from request when certificate exists
    """
    ca_path = tmp_path
    ca_name = "test_ca"
    certp = "{}/{}/certs/{}.crt".format(
        ca_path, ca_name, tls_test_data["create_ca"]["CN"]
    )
    ret = 'Created Certificate for "{}": "{}"'.format(
        tls_test_data["create_ca"]["CN"], certp
    )
    mock_opt = MagicMock(return_value=ca_path)
    mock_ret = MagicMock(return_value=0)
    mock_pgt = MagicMock(return_value=False)
    with patch.dict(
        tls.__salt__,
        {
            "config.option": mock_opt,
            "cmd.retcode": mock_ret,
            "pillar.get": mock_pgt,
        },
    ), patch.dict(tls.__opts__, {"hash_type": "sha256", "cachedir": ca_path}), patch(
        "salt.modules.tls.maybe_fix_ssl_version", MagicMock(return_value=True)
    ):
        tls.create_ca(ca_name)
        tls.create_csr(ca_name)
        tls.create_ca_signed_cert(ca_name, tls_test_data["create_ca"]["CN"])
        assert (
            tls.create_ca_signed_cert(
                ca_name, tls_test_data["create_ca"]["CN"], replace=True
            )
            == ret
        )


def test_create_pkcs12(tmp_path, tls_test_data):
    """
    Test creating pkcs12
    """
    ca_path = tmp_path
    ca_name = "test_ca"
    certp = "{}/{}/certs/{}.p12".format(
        ca_path, ca_name, tls_test_data["create_ca"]["CN"]
    )
    ret = 'Created PKCS#12 Certificate for "{}": "{}"'.format(
        tls_test_data["create_ca"]["CN"], certp
    )
    mock_opt = MagicMock(return_value=ca_path)
    mock_ret = MagicMock(return_value=0)
    mock_pgt = MagicMock(return_value=False)
    with patch.dict(
        tls.__salt__,
        {
            "config.option": mock_opt,
            "cmd.retcode": mock_ret,
            "pillar.get": mock_pgt,
        },
    ), patch.dict(tls.__opts__, {"hash_type": "sha256", "cachedir": ca_path}), patch(
        "salt.modules.tls.maybe_fix_ssl_version", MagicMock(return_value=True)
    ):
        tls.create_ca(ca_name)
        tls.create_csr(ca_name, **tls_test_data["create_ca"])
        tls.create_ca_signed_cert(ca_name, tls_test_data["create_ca"]["CN"])
        assert (
            tls.create_pkcs12(ca_name, tls_test_data["create_ca"]["CN"], "password")
            == ret
        )


def test_recreate_pkcs12(tmp_path, tls_test_data):
    """
    Test creating pkcs12 when it already exists
    """
    ca_path = tmp_path
    ca_name = "test_ca"
    certp = "{}/{}/certs/{}.p12".format(
        ca_path, ca_name, tls_test_data["create_ca"]["CN"]
    )
    ret = 'Created PKCS#12 Certificate for "{}": "{}"'.format(
        tls_test_data["create_ca"]["CN"], certp
    )
    mock_opt = MagicMock(return_value=ca_path)
    mock_ret = MagicMock(return_value=0)
    mock_pgt = MagicMock(return_value=False)
    with patch.dict(
        tls.__salt__,
        {
            "config.option": mock_opt,
            "cmd.retcode": mock_ret,
            "pillar.get": mock_pgt,
        },
    ), patch.dict(
        tls.__opts__, {"hash_type": "sha256", "cachedir": ca_path}
    ), patch.dict(
        tls_test_data["create_ca"], {"replace": True}
    ), patch(
        "salt.modules.tls.maybe_fix_ssl_version", MagicMock(return_value=True)
    ):
        tls.create_ca(ca_name)
        tls.create_csr(ca_name)
        tls.create_ca_signed_cert(ca_name, tls_test_data["create_ca"]["CN"])
        tls.create_pkcs12(ca_name, tls_test_data["create_ca"]["CN"], "password")
        assert (
            tls.create_pkcs12(
                ca_name, tls_test_data["create_ca"]["CN"], "password", replace=True
            )
            == ret
        )


def test_pyOpenSSL_version():
    """
    Test extension logic with different pyOpenSSL versions
    """
    pillarval = {"csr": {"extendedKeyUsage": "serverAuth"}}
    mock_pgt = MagicMock(return_value=pillarval)
    with patch.dict(
        tls.__dict__,
        {"OpenSSL_version": Version("0.1.1"), "X509_EXT_ENABLED": False},
    ):
        assert tls.__virtual__() == (
            False,
            "PyOpenSSL version 0.10 or later must be installed before this module can be used.",
        )
        with patch.dict(tls.__salt__, {"pillar.get": mock_pgt}):
            for thing in ("server", "client"):
                with pytest.raises(AssertionError):
                    tls.get_extensions(thing)
    with patch.dict(
        tls.__dict__,
        {"OpenSSL_version": Version("0.14.1"), "X509_EXT_ENABLED": True},
    ):
        assert tls.__virtual__()
        with patch.dict(tls.__salt__, {"pillar.get": mock_pgt}):
            assert tls.get_extensions("server") == pillarval
            assert tls.get_extensions("client") == pillarval
    with patch.dict(
        tls.__dict__,
        {"OpenSSL_version": Version("0.15.1"), "X509_EXT_ENABLED": True},
    ):
        assert tls.__virtual__()
        with patch.dict(tls.__salt__, {"pillar.get": mock_pgt}):
            assert tls.get_extensions("server") == pillarval
            assert tls.get_extensions("client") == pillarval


def test_pyOpenSSL_version_destructive(tmp_path, tls_test_data):
    """
    Test extension logic with different pyOpenSSL versions
    """
    ca_path = tmp_path
    pillarval = {"csr": {"extendedKeyUsage": "serverAuth"}}
    mock_pgt = MagicMock(return_value=pillarval)
    ca_name = "test_ca"
    certp = "{0}/{1}/{1}_ca_cert.crt".format(ca_path, ca_name)
    certk = "{0}/{1}/{1}_ca_cert.key".format(ca_path, ca_name)
    ret = 'Created Private Key: "{}" Created CA "{}": "{}"'.format(
        certk, ca_name, certp
    )
    mock_opt = MagicMock(return_value=ca_path)
    mock_ret = MagicMock(return_value=0)
    with patch.dict(tls.__salt__, {"config.option": mock_opt, "cmd.retcode": mock_ret}):
        with patch.dict(tls.__opts__, {"hash_type": "sha256", "cachedir": ca_path}):
            with patch.dict(tls_test_data["create_ca"], {"replace": True}):
                with patch.dict(
                    tls.__dict__,
                    {
                        "OpenSSL_version": Version("0.1.1"),
                        "X509_EXT_ENABLED": False,
                    },
                ):
                    assert (
                        tls.create_ca(
                            ca_name,
                            days=365,
                            fixmode=False,
                            **tls_test_data["create_ca"],
                        )
                        == ret
                    )
                with patch.dict(
                    tls.__dict__,
                    {
                        "OpenSSL_version": Version("0.14.1"),
                        "X509_EXT_ENABLED": True,
                    },
                ):
                    assert (
                        tls.create_ca(
                            ca_name,
                            days=365,
                            fixmode=False,
                            **tls_test_data["create_ca"],
                        )
                        == ret
                    )
                with patch.dict(
                    tls.__dict__,
                    {
                        "OpenSSL_version": Version("0.15.1"),
                        "X509_EXT_ENABLED": True,
                    },
                ):
                    assert (
                        tls.create_ca(
                            ca_name,
                            days=365,
                            fixmode=False,
                            **tls_test_data["create_ca"],
                        )
                        == ret
                    )

    certp = "{}/{}/certs/{}.csr".format(
        ca_path, ca_name, tls_test_data["create_ca"]["CN"]
    )
    certk = "{}/{}/certs/{}.key".format(
        ca_path, ca_name, tls_test_data["create_ca"]["CN"]
    )
    ret = 'Created Private Key: "{}" Created CSR for "{}": "{}"'.format(
        certk, tls_test_data["create_ca"]["CN"], certp
    )
    with patch.dict(
        tls.__salt__,
        {
            "config.option": mock_opt,
            "cmd.retcode": mock_ret,
            "pillar.get": mock_pgt,
        },
    ):
        with patch.dict(tls.__opts__, {"hash_type": "sha256", "cachedir": ca_path}):
            with patch.dict(
                tls_test_data["create_ca"],
                {"subjectAltName": "DNS:foo.bar", "replace": True},
            ):
                with patch.dict(
                    tls.__dict__,
                    {
                        "OpenSSL_version": Version("0.1.1"),
                        "X509_EXT_ENABLED": False,
                    },
                ):
                    tls.create_ca(ca_name)
                    tls.create_csr(ca_name)
                    with pytest.raises(ValueError):
                        tls.create_csr(ca_name, **tls_test_data["create_ca"])
                with patch.dict(
                    tls.__dict__,
                    {
                        "OpenSSL_version": Version("0.14.1"),
                        "X509_EXT_ENABLED": True,
                    },
                ):
                    tls.create_ca(ca_name)
                    tls.create_csr(ca_name)
                    assert tls.create_csr(ca_name, **tls_test_data["create_ca"]) == ret
                with patch.dict(
                    tls.__dict__,
                    {
                        "OpenSSL_version": Version("0.15.1"),
                        "X509_EXT_ENABLED": True,
                    },
                ):
                    tls.create_ca(ca_name)
                    tls.create_csr(ca_name)
                    assert tls.create_csr(ca_name, **tls_test_data["create_ca"]) == ret


def test_get_expiration_date(tls_test_data):
    with patch("salt.utils.files.fopen", mock_open(read_data=tls_test_data["ca_cert"])):
        assert tls.get_expiration_date("/path/to/cert") == "2016-05-04"
    with patch("salt.utils.files.fopen", mock_open(read_data=tls_test_data["ca_cert"])):
        assert (
            tls.get_expiration_date("/path/to/cert", date_format="%d/%m/%y")
            == "04/05/16"
        )



"""
PyInstaller runtime hook to patch salt.utils.vt.Terminal
"""
import logging

import salt.utils.vt
from salt.utils.pyinstaller.rthooks._overrides import PyinstallerTerminal

log = logging.getLogger(__name__)
# Patch salt.utils.vt.Terminal when running within a pyinstalled bundled package
salt.utils.vt.Terminal = PyinstallerTerminal

log.debug("Replaced 'salt.utils.vt.Terminal' with 'PyinstallerTerminal'")



r"""
A salt util for modifying firewall settings.

.. versionadded:: 2018.3.4
.. versionadded:: 2019.2.0

This util allows you to modify firewall settings in the local group policy in
addition to the normal firewall settings. Parameters are taken from the
netsh advfirewall prompt.

.. note::
    More information can be found in the advfirewall context in netsh. This can
    be access by opening a netsh prompt. At a command prompt type the following:

    c:\>netsh
    netsh>advfirewall
    netsh advfirewall>set help
    netsh advfirewall>set domain help

Usage:

.. code-block:: python

    import salt.utils.win_lgpo_netsh

    # Get the inbound/outbound firewall settings for connections on the
    # local domain profile
    salt.utils.win_lgpo_netsh.get_settings(profile='domain',
                                           section='firewallpolicy')

    # Get the inbound/outbound firewall settings for connections on the
    # domain profile as defined by local group policy
    salt.utils.win_lgpo_netsh.get_settings(profile='domain',
                                           section='firewallpolicy',
                                           store='lgpo')

    # Get all firewall settings for connections on the domain profile
    salt.utils.win_lgpo_netsh.get_all_settings(profile='domain')

    # Get all firewall settings for connections on the domain profile as
    # defined by local group policy
    salt.utils.win_lgpo_netsh.get_all_settings(profile='domain', store='lgpo')

    # Get all firewall settings for all profiles
    salt.utils.win_lgpo_netsh.get_all_settings()

    # Get all firewall settings for all profiles as defined by local group
    # policy
    salt.utils.win_lgpo_netsh.get_all_settings(store='lgpo')

    # Set the inbound setting for the domain profile to block inbound
    # connections
    salt.utils.win_lgpo_netsh.set_firewall_settings(profile='domain',
                                                    inbound='blockinbound')

    # Set the outbound setting for the domain profile to allow outbound
    # connections
    salt.utils.win_lgpo_netsh.set_firewall_settings(profile='domain',
                                                    outbound='allowoutbound')

    # Set inbound/outbound settings for the domain profile in the group
    # policy to block inbound and allow outbound
    salt.utils.win_lgpo_netsh.set_firewall_settings(profile='domain',
                                                    inbound='blockinbound',
                                                    outbound='allowoutbound',
                                                    store='lgpo')
"""

import logging
import os
import re
import socket
import tempfile
from textwrap import dedent

import salt.modules.cmdmod
from salt.exceptions import CommandExecutionError

log = logging.getLogger(__name__)
__hostname__ = socket.gethostname()
__virtualname__ = "netsh"


# Although utils are often directly imported, it is also possible to use the
# loader.
def __virtual__():
    """
    Only load if on a Windows system
    """
    if not salt.utils.platform.is_windows():
        return False, "This utility only available on Windows"

    return __virtualname__


def _netsh_file(content):
    """
    helper function to get the results of ``netsh -f content.txt``

    Running ``netsh`` will drop you into a ``netsh`` prompt where you can issue
    ``netsh`` commands. You can put a series of commands in an external file and
    run them as if from a ``netsh`` prompt using the ``-f`` switch. That's what
    this function does.

    Args:

        content (str):
            The contents of the file that will be run by the ``netsh -f``
            command

    Returns:
        str: The text returned by the netsh command
    """
    with tempfile.NamedTemporaryFile(
        mode="w", prefix="salt-", suffix=".netsh", delete=False, encoding="utf-8"
    ) as fp:
        fp.write(content)
    try:
        log.debug("%s:\n%s", fp.name, content)
        return salt.modules.cmdmod.run("netsh -f {}".format(fp.name), python_shell=True)
    finally:
        os.remove(fp.name)


def _netsh_command(command, store):
    if store.lower() not in ("local", "lgpo"):
        raise ValueError("Incorrect store: {}".format(store))
    # set the store for local or lgpo
    if store.lower() == "local":
        netsh_script = dedent(
            """\
            advfirewall
            set store local
            {}
        """.format(
                command
            )
        )
    else:
        netsh_script = dedent(
            """\
            advfirewall
            set store gpo = {}
            {}
        """.format(
                __hostname__, command
            )
        )
    return _netsh_file(content=netsh_script).splitlines()


def get_settings(profile, section, store="local"):
    """
    Get the firewall property from the specified profile in the specified store
    as returned by ``netsh advfirewall``.

    Args:

        profile (str):
            The firewall profile to query. Valid options are:

            - domain
            - public
            - private

        section (str):
            The property to query within the selected profile. Valid options
            are:

            - firewallpolicy : inbound/outbound behavior
            - logging : firewall logging settings
            - settings : firewall properties
            - state : firewalls state (on | off)

        store (str):
            The store to use. This is either the local firewall policy or the
            policy defined by local group policy. Valid options are:

            - lgpo
            - local

            Default is ``local``

    Returns:
        dict: A dictionary containing the properties for the specified profile

    Raises:
        CommandExecutionError: If an error occurs
        ValueError: If the parameters are incorrect
    """
    # validate input
    if profile.lower() not in ("domain", "public", "private"):
        raise ValueError("Incorrect profile: {}".format(profile))
    if section.lower() not in ("state", "firewallpolicy", "settings", "logging"):
        raise ValueError("Incorrect section: {}".format(section))
    if store.lower() not in ("local", "lgpo"):
        raise ValueError("Incorrect store: {}".format(store))
    command = "show {}profile {}".format(profile, section)
    # run it
    results = _netsh_command(command=command, store=store)
    # sample output:
    # Domain Profile Settings:
    # ----------------------------------------------------------------------
    # LocalFirewallRules                    N/A (GPO-store only)
    # LocalConSecRules                      N/A (GPO-store only)
    # InboundUserNotification               Disable
    # RemoteManagement                      Disable
    # UnicastResponseToMulticast            Enable

    # if it's less than 3 lines it failed
    if len(results) < 3:
        raise CommandExecutionError("Invalid results: {}".format(results))
    ret = {}
    # Skip the first 2 lines. Add everything else to a dictionary
    for line in results[3:]:
        ret.update(dict(list(zip(*[iter(re.split(r"\s{2,}", line))] * 2))))

    # Remove spaces from the values so that `Not Configured` is detected
    # correctly
    for item in ret:
        ret[item] = ret[item].replace(" ", "")

    # special handling for firewallpolicy
    if section == "firewallpolicy":
        inbound, outbound = ret["Firewall Policy"].split(",")
        return {"Inbound": inbound, "Outbound": outbound}

    return ret


def get_all_settings(profile, store="local"):
    """
    Gets all the properties for the specified profile in the specified store

    Args:

        profile (str):
            The firewall profile to query. Valid options are:

            - domain
            - public
            - private

        store (str):
            The store to use. This is either the local firewall policy or the
            policy defined by local group policy. Valid options are:

            - lgpo
            - local

            Default is ``local``

    Returns:
        dict: A dictionary containing the specified settings
    """
    ret = dict()
    ret.update(get_settings(profile=profile, section="state", store=store))
    ret.update(get_settings(profile=profile, section="firewallpolicy", store=store))
    ret.update(get_settings(profile=profile, section="settings", store=store))
    ret.update(get_settings(profile=profile, section="logging", store=store))
    return ret


def get_all_profiles(store="local"):
    """
    Gets all properties for all profiles in the specified store

    Args:

        store (str):
            The store to use. This is either the local firewall policy or the
            policy defined by local group policy. Valid options are:

            - lgpo
            - local

            Default is ``local``

    Returns:
        dict: A dictionary containing the specified settings for each profile
    """
    return {
        "Domain Profile": get_all_settings(profile="domain", store=store),
        "Private Profile": get_all_settings(profile="private", store=store),
        "Public Profile": get_all_settings(profile="public", store=store),
    }


def set_firewall_settings(profile, inbound=None, outbound=None, store="local"):
    """
    Set the firewall inbound/outbound settings for the specified profile and
    store

    Args:

        profile (str):
            The firewall profile to configure. Valid options are:

            - domain
            - public
            - private

        inbound (str):
            The inbound setting. If ``None`` is passed, the setting will remain
            unchanged. Valid values are:

            - blockinbound
            - blockinboundalways
            - allowinbound
            - notconfigured

            Default is ``None``

        outbound (str):
            The outbound setting. If ``None`` is passed, the setting will remain
            unchanged. Valid values are:

            - allowoutbound
            - blockoutbound
            - notconfigured

            Default is ``None``

        store (str):
            The store to use. This is either the local firewall policy or the
            policy defined by local group policy. Valid options are:

            - lgpo
            - local

            Default is ``local``

    Returns:
        bool: ``True`` if successful

    Raises:
        CommandExecutionError: If an error occurs
        ValueError: If the parameters are incorrect
    """
    # Input validation
    if profile.lower() not in ("domain", "public", "private"):
        raise ValueError("Incorrect profile: {}".format(profile))
    if inbound and inbound.lower() not in (
        "blockinbound",
        "blockinboundalways",
        "allowinbound",
        "notconfigured",
    ):
        raise ValueError("Incorrect inbound value: {}".format(inbound))
    if outbound and outbound.lower() not in (
        "allowoutbound",
        "blockoutbound",
        "notconfigured",
    ):
        raise ValueError("Incorrect outbound value: {}".format(outbound))
    if not inbound and not outbound:
        raise ValueError("Must set inbound or outbound")

    # You have to specify inbound and outbound setting at the same time
    # If you're only specifying one, you have to get the current setting for the
    # other
    if not inbound or not outbound:
        ret = get_settings(profile=profile, section="firewallpolicy", store=store)
        if not inbound:
            inbound = ret["Inbound"]
        if not outbound:
            outbound = ret["Outbound"]

    command = "set {}profile firewallpolicy {},{}".format(profile, inbound, outbound)

    results = _netsh_command(command=command, store=store)

    if results:
        raise CommandExecutionError("An error occurred: {}".format(results))

    return True


def set_logging_settings(profile, setting, value, store="local"):
    """
    Configure logging settings for the Windows firewall.

    Args:

        profile (str):
            The firewall profile to configure. Valid options are:

            - domain
            - public
            - private

        setting (str):
            The logging setting to configure. Valid options are:

            - allowedconnections
            - droppedconnections
            - filename
            - maxfilesize

        value (str):
            The value to apply to the setting. Valid values are dependent upon
            the setting being configured. Valid options are:

            allowedconnections:

                - enable
                - disable
                - notconfigured

            droppedconnections:

                - enable
                - disable
                - notconfigured

            filename:

                - Full path and name of the firewall log file
                - notconfigured

            maxfilesize:

                - 1 - 32767 (Kb)
                - notconfigured

        store (str):
            The store to use. This is either the local firewall policy or the
            policy defined by local group policy. Valid options are:

            - lgpo
            - local

            Default is ``local``

    Returns:
        bool: ``True`` if successful

    Raises:
        CommandExecutionError: If an error occurs
        ValueError: If the parameters are incorrect
    """
    # Input validation
    if profile.lower() not in ("domain", "public", "private"):
        raise ValueError("Incorrect profile: {}".format(profile))
    if setting.lower() not in (
        "allowedconnections",
        "droppedconnections",
        "filename",
        "maxfilesize",
    ):
        raise ValueError("Incorrect setting: {}".format(setting))
    if setting.lower() in ("allowedconnections", "droppedconnections"):
        if value.lower() not in ("enable", "disable", "notconfigured"):
            raise ValueError("Incorrect value: {}".format(value))
    # TODO: Consider adding something like the following to validate filename
    # https://stackoverflow.com/questions/9532499/check-whether-a-path-is-valid-in-python-without-creating-a-file-at-the-paths-ta
    if setting.lower() == "maxfilesize":
        if value.lower() != "notconfigured":
            # Must be a number between 1 and 32767
            try:
                int(value)
            except ValueError:
                raise ValueError("Incorrect value: {}".format(value))
            if not 1 <= int(value) <= 32767:
                raise ValueError("Incorrect value: {}".format(value))
    # Run the command
    command = "set {}profile logging {} {}".format(profile, setting, value)
    results = _netsh_command(command=command, store=store)

    # A successful run should return an empty list
    if results:
        raise CommandExecutionError("An error occurred: {}".format(results))

    return True


def set_settings(profile, setting, value, store="local"):
    """
    Configure firewall settings.

    Args:

        profile (str):
            The firewall profile to configure. Valid options are:

            - domain
            - public
            - private

        setting (str):
            The firewall setting to configure. Valid options are:

            - localfirewallrules
            - localconsecrules
            - inboundusernotification
            - remotemanagement
            - unicastresponsetomulticast

        value (str):
            The value to apply to the setting. Valid options are

            - enable
            - disable
            - notconfigured

        store (str):
            The store to use. This is either the local firewall policy or the
            policy defined by local group policy. Valid options are:

            - lgpo
            - local

            Default is ``local``

    Returns:
        bool: ``True`` if successful

    Raises:
        CommandExecutionError: If an error occurs
        ValueError: If the parameters are incorrect
    """
    # Input validation
    if profile.lower() not in ("domain", "public", "private"):
        raise ValueError("Incorrect profile: {}".format(profile))
    if setting.lower() not in (
        "localfirewallrules",
        "localconsecrules",
        "inboundusernotification",
        "remotemanagement",
        "unicastresponsetomulticast",
    ):
        raise ValueError("Incorrect setting: {}".format(setting))
    if value.lower() not in ("enable", "disable", "notconfigured"):
        raise ValueError("Incorrect value: {}".format(value))

    # Run the command
    command = "set {}profile settings {} {}".format(profile, setting, value)
    results = _netsh_command(command=command, store=store)

    # A successful run should return an empty list
    if results:
        raise CommandExecutionError("An error occurred: {}".format(results))

    return True


def set_state(profile, state, store="local"):
    """
    Configure the firewall state.

    Args:

        profile (str):
            The firewall profile to configure. Valid options are:

            - domain
            - public
            - private

        state (str):
            The firewall state. Valid options are:

            - on
            - off
            - notconfigured

        store (str):
            The store to use. This is either the local firewall policy or the
            policy defined by local group policy. Valid options are:

            - lgpo
            - local

            Default is ``local``

    Returns:
        bool: ``True`` if successful

    Raises:
        CommandExecutionError: If an error occurs
        ValueError: If the parameters are incorrect
    """
    # Input validation
    if profile.lower() not in ("domain", "public", "private"):
        raise ValueError("Incorrect profile: {}".format(profile))
    if state.lower() not in ("on", "off", "notconfigured"):
        raise ValueError("Incorrect state: {}".format(state))

    # Run the command
    command = "set {}profile state {}".format(profile, state)
    results = _netsh_command(command=command, store=store)

    # A successful run should return an empty list
    if results:
        raise CommandExecutionError("An error occurred: {}".format(results))

    return True



"""
Parses roster entries out of Host directives from SSH config

.. code-block:: bash

    salt-ssh --roster sshconfig '*' -r "echo hi"
"""

import collections
import fnmatch
import logging
import os
import re

import salt.utils.files
import salt.utils.stringutils

log = logging.getLogger(__name__)

_SSHConfRegex = collections.namedtuple("_SSHConfRegex", ["target_field", "pattern"])
_ROSTER_FIELDS = (
    _SSHConfRegex(target_field="user", pattern=r"\s+User (.*)"),
    _SSHConfRegex(target_field="port", pattern=r"\s+Port (.*)"),
    _SSHConfRegex(target_field="priv", pattern=r"\s+IdentityFile (.*)"),
)


def _get_ssh_config_file(opts):
    """
    :return: Path to the .ssh/config file - usually <home>/.ssh/config
    """
    ssh_config_file = opts.get("ssh_config_file")
    if not os.path.isfile(ssh_config_file):
        raise OSError("Cannot find SSH config file")
    if not os.access(ssh_config_file, os.R_OK):
        raise OSError("Cannot access SSH config file: {}".format(ssh_config_file))
    return ssh_config_file


def parse_ssh_config(lines):
    """
    Parses lines from the SSH config to create roster targets.

    :param lines: Individual lines from the ssh config file
    :return: Dictionary of targets in similar style to the flat roster
    """
    # transform the list of individual lines into a list of sublists where each
    # sublist represents a single Host definition
    hosts = []
    for line in lines:
        line = salt.utils.stringutils.to_unicode(line)
        if not line or line.startswith("#"):
            continue
        elif line.startswith("Host "):
            hosts.append([])
        hosts[-1].append(line)

    # construct a dictionary of Host names to mapped roster properties
    targets = collections.OrderedDict()
    for host_data in hosts:
        target = collections.OrderedDict()
        hostnames = host_data[0].split()[1:]
        for line in host_data[1:]:
            for field in _ROSTER_FIELDS:
                match = re.match(field.pattern, line)
                if match:
                    target[field.target_field] = match.group(1).strip()
        for hostname in hostnames:
            targets[hostname] = target

    # apply matching for glob hosts
    wildcard_targets = []
    non_wildcard_targets = []
    for target in targets.keys():
        if "*" in target or "?" in target:
            wildcard_targets.append(target)
        else:
            non_wildcard_targets.append(target)
    for pattern in wildcard_targets:
        for candidate in non_wildcard_targets:
            if fnmatch.fnmatch(candidate, pattern):
                targets[candidate].update(targets[pattern])
        del targets[pattern]

    # finally, update the 'host' to refer to its declaration in the SSH config
    # so that its connection parameters can be utilized
    for target in targets:
        targets[target]["host"] = target
    return targets


def targets(tgt, tgt_type="glob", **kwargs):
    """
    Return the targets from the flat yaml file, checks opts for location but
    defaults to /etc/salt/roster
    """
    ssh_config_file = _get_ssh_config_file(__opts__)
    with salt.utils.files.fopen(ssh_config_file, "r") as fp:
        all_minions = parse_ssh_config([line.rstrip() for line in fp])
    rmatcher = RosterMatcher(all_minions, tgt, tgt_type)
    matched = rmatcher.targets()
    return matched


class RosterMatcher:
    """
    Matcher for the roster data structure
    """

    def __init__(self, raw, tgt, tgt_type):
        self.tgt = tgt
        self.tgt_type = tgt_type
        self.raw = raw

    def targets(self):
        """
        Execute the correct tgt_type routine and return
        """
        try:
            return getattr(self, "ret_{}_minions".format(self.tgt_type))()
        except AttributeError:
            return {}

    def ret_glob_minions(self):
        """
        Return minions that match via glob
        """
        minions = {}
        for minion in self.raw:
            if fnmatch.fnmatch(minion, self.tgt):
                data = self.get_data(minion)
                if data:
                    minions[minion] = data
        return minions

    def get_data(self, minion):
        """
        Return the configured ip
        """
        if isinstance(self.raw[minion], str):
            return {"host": self.raw[minion]}
        if isinstance(self.raw[minion], dict):
            return self.raw[minion]
        return False



"""
    :codeauthor: Rupesh Tare <rupesht@saltstack.com>
"""

import os

import pytest

import salt.modules.debconfmod as debconfmod
from tests.support.mock import MagicMock, patch


@pytest.fixture
def configure_loader_modules():
    return {debconfmod: {}}


def test_get_selections():
    """
    Test for Answers to debconf questions for all packages
    """
    mock = MagicMock(return_value=[])
    with patch.dict(debconfmod.__salt__, {"cmd.run_stdout": mock}):
        with patch.object(debconfmod, "_unpack_lines", mock):
            assert debconfmod.get_selections(False) == {}


def test_show():
    """
    Test for Answers to debconf questions for a package
    """
    mock = MagicMock(return_value={})
    with patch.object(debconfmod, "get_selections", mock):
        assert debconfmod.show("name") is None


def test_set_():
    """
    Test for Set answers to debconf questions for a package.
    """
    mock = MagicMock(return_value=None)
    with patch.object(os, "write", mock):
        with patch.object(os, "close", mock):
            with patch.object(debconfmod, "_set_file", mock):
                with patch.object(os, "unlink", mock):
                    assert debconfmod.set_("package", "question", "type", "value")


def test_set_template():
    """
    Test for Set answers to debconf questions from a template.
    """
    mock = MagicMock(return_value="A")
    with patch.dict(debconfmod.__salt__, {"cp.get_template": mock}):
        with patch.object(debconfmod, "set_file", mock):
            assert (
                debconfmod.set_template(
                    "path", "template", "context", "defaults", "saltenv"
                )
                == "A"
            )


def test_set_file():
    """
    Test for Set answers to debconf questions from a file.
    """
    mock = MagicMock(return_value="A")
    with patch.dict(debconfmod.__salt__, {"cp.cache_file": mock}):
        mock = MagicMock(return_value=None)
        with patch.object(debconfmod, "_set_file", mock):
            assert debconfmod.set_file("path")

    mock = MagicMock(return_value=False)
    with patch.dict(debconfmod.__salt__, {"cp.cache_file": mock}):
        mock = MagicMock(return_value=None)
        with patch.object(debconfmod, "_set_file", mock):
            assert not debconfmod.set_file("path")
